package com.arms.api.issue.almapi.service;

import static com.arms.api.util.alm.지라유틸.*;

import com.arms.api.issue.almapi.model.dto.*;
import com.arms.api.issue.almapi.model.entity.AlmIssueEntity;
import com.arms.api.issue.almapi.model.vo.지라이슈_데이터;
import com.arms.api.issue.almapi.repository.지라이슈_저장소;
import com.arms.api.issue.priority.model.IssuePriorityDTO;
import com.arms.api.issue.resolution.model.이슈해결책_데이터;
import com.arms.api.issue.status.dto.IssueStatusDTO;
import com.arms.api.issue.type.dto.IssueTypeDTO;
import com.arms.api.project.dto.ProjectDTO;
import com.arms.api.serverinfo.model.서버정보_데이터;
import com.arms.api.util.alm.지라API_정보;
import com.arms.api.util.alm.지라유틸;
import com.arms.egovframework.javaservice.esframework.erros.response.ErrorCode;
import com.arms.api.util.errors.ErrorLogUtil;
import com.atlassian.jira.rest.client.api.JiraRestClient;
import com.atlassian.jira.rest.client.api.RestClientException;
import com.atlassian.jira.rest.client.api.domain.*;
import com.atlassian.jira.rest.client.api.domain.input.IssueInput;
import com.atlassian.jira.rest.client.api.domain.input.IssueInputBuilder;
import com.atlassian.jira.rest.client.api.domain.input.TransitionInput;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.joda.JodaModule;
import io.atlassian.util.concurrent.Promise;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpStatus;
import org.joda.time.DateTime;
import org.springframework.stereotype.Component;
import org.springframework.web.reactive.function.client.WebClient;

import java.util.*;

@Slf4j
@Component
@AllArgsConstructor
public class 온프레미스_지라_이슈전략 implements 이슈전략 {

    private final 지라유틸 지라유틸;
    private final 지라API_정보 지라API_정보;
    private final 지라이슈_저장소 지라이슈_저장소;

    @Override
    public List<지라이슈_데이터> getIssueList(서버정보_데이터 서버정보, String projectKeyOrId) {

        JiraRestClient restClient;
        try {
            restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                    서버정보.getUserId(),
                    서버정보.getPasswordOrToken());
        }
        catch (Exception e) {
            String errorMessage = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라("+ 서버정보.getUri() +") :: 프로젝트("+ projectKeyOrId+ ") :: 전체이슈_목록_가져오기에 실패하였습니다.");
            throw new IllegalArgumentException(ErrorCode.이슈_조회_오류.getErrorMsg() + " :: " + errorMessage);
        }

        String 조회할_프로젝트 = "project = " + projectKeyOrId;

        int startAt = 0;
        final int maxResults = 지라API_정보.getParameter().getMaxResults();
        Set<String> 필드 = new HashSet<>(Arrays.asList("*all")); // 검색 필드

        List<지라이슈_데이터> issueList = new ArrayList<>();
        while (true) {
            SearchResult searchResult = restClient.getSearchClient()
                    .searchJql(조회할_프로젝트, maxResults, startAt, 필드)
                    .claim();

            if (searchResult == null || searchResult.getIssues() == null) {
                String errorMessage = String.format("온프레미스 지라[%s] :: 프로젝트[%s] :: getIssueList searchResult is null.", 서버정보.getUri(), projectKeyOrId);
                log.error(errorMessage);

                throw new IllegalArgumentException(errorMessage);
            }

            searchResult.getIssues().forEach(issue -> issueList.add(지라이슈_데이터로_변환(issue)));

            if (checkIfLast(issueList.size(), searchResult.getTotal())) {
                break;
            }

            startAt += maxResults;
        }

        return issueList;
    }

    private boolean checkIfLast(int totalSize, int currentSize) {
        return totalSize == currentSize;
    }

    /* ***
     * 수정사항: null 체크하여 에러 처리 필요
     *** */
    @Override
    public 지라이슈_데이터 createIssue(서버정보_데이터 서버정보, AlmIssueCreationDTO almIssueCreationDTO) {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        IssueCreationFieldsDTO fieldsDTO = almIssueCreationDTO.getFields();
        if (fieldsDTO == null) {
            String errorMessage = String.format("%s[%s] :: 이슈 생성 필드 데이터가 존재 하지 않습니다.",
                    서버정보.getType(), 서버정보.getUri());
            log.error(errorMessage);
            throw new IllegalArgumentException(ErrorCode.요청본문_오류체크.getErrorMsg() + " :: " + errorMessage);
        }

        String projectKeyOrId = null;
        Long issueTypeId = null;
        String summary = null;

        if (fieldsDTO.getProject() != null && StringUtils.isNotBlank(fieldsDTO.getProject().getKey())) {
            projectKeyOrId = fieldsDTO.getProject().getKey();
        }

        if (fieldsDTO.getIssuetype() != null && StringUtils.isNotEmpty(fieldsDTO.getIssuetype().getId())) {
            issueTypeId = Long.valueOf(fieldsDTO.getIssuetype().getId());
        }

        if (StringUtils.isEmpty(projectKeyOrId) || issueTypeId == null) {
            String errorMessage = String.format("%s[%s] :: 이슈 생성 필드 확인에 필요한 프로젝트 아이디, 이슈유형 아이디가 존재하지 않습니다.",
                    서버정보.getType(), 서버정보.getUri());
            throw new IllegalArgumentException(errorMessage);
        }

        if (StringUtils.isNotEmpty(fieldsDTO.getSummary())) {
            summary = fieldsDTO.getSummary();
        }

        IssueInputBuilder issueInputBuilder = new IssueInputBuilder(projectKeyOrId, issueTypeId, summary);

        if (StringUtils.isNotEmpty(fieldsDTO.getDescription())) {
            issueInputBuilder.setDescription(fieldsDTO.getDescription());
        }

        if (fieldsDTO.getReporter() != null && StringUtils.isNotEmpty(fieldsDTO.getReporter().getAccountId())) {
            issueInputBuilder.setReporterName(fieldsDTO.getReporter().getAccountId());
        }

        if (fieldsDTO.getPriority() != null && StringUtils.isNotEmpty(fieldsDTO.getPriority().getId())) {
            issueInputBuilder.setPriorityId(Long.valueOf(fieldsDTO.getPriority().getId()));
        }

        if (fieldsDTO.getDueDate() != null) {
            DateTime dateTime = new DateTime(fieldsDTO.getDueDate());
            issueInputBuilder.setDueDate(dateTime);
        }

        IssueInput issueInput = issueInputBuilder.build();

        BasicIssue basicIssue = restClient.getIssueClient().createIssue(issueInput).claim();

        if (basicIssue == null) {
            String errorMessage = String.format("%s[%s], 프로젝트[%s], 이슈유형[%s], 생성 필드 :: %s 생성된 이슈가 존재 하지 않습니다.",
                                                서버정보.getType(), 서버정보.getUri(), projectKeyOrId, issueTypeId, issueInput.toString());
            log.error(errorMessage);
            throw new IllegalArgumentException(ErrorCode.이슈생성_오류.getErrorMsg() + errorMessage);
        }

        지라이슈_데이터 returnIssueDTO = new 지라이슈_데이터();
        returnIssueDTO.setId(basicIssue.getId().toString());
        returnIssueDTO.setKey(basicIssue.getKey());
        returnIssueDTO.setSelf(basicIssue.getSelf().toString());

        return returnIssueDTO;
    }

    @Override
    public Map<String, Object> 이슈_수정하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디, AlmIssueCreationDTO almIssueCreationDTO) {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        Map<String, Object> 결과 = new HashMap<>();
        IssueCreationFieldsDTO 필드_데이터 = almIssueCreationDTO.getFields();

        IssueInputBuilder 입력_생성 = new IssueInputBuilder();

        if (필드_데이터.getSummary() != null) {
            입력_생성.setSummary(필드_데이터.getSummary());
        }

        if (필드_데이터.getDescription() != null) {
            입력_생성.setDescription(필드_데이터.getDescription());
        }

        if (필드_데이터.getLabels() != null) {
            입력_생성.setFieldValue("labels", 필드_데이터.getLabels());
        }

        IssueInput 수정_데이터 = 입력_생성.build();
        try {
            Promise<Void> 수정결과 = restClient.getIssueClient().updateIssue(이슈_키_또는_아이디, 수정_데이터);
            수정결과.claim();

            결과.put("success", true);
            결과.put("message", "이슈 수정 성공");
        }
        catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() +") ::  생성 필드 :: "
                            + 필드_데이터.toString() + ", :: 이슈_수정하기 중 오류 :: " + 이슈_키_또는_아이디);
            log.error(에러로그);

            결과.put("success", false);
            결과.put("message", 에러로그);
        }

        return 결과;
    }

    @Override
    public Map<String, Object> 이슈_상태_변경하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디, String 상태_아이디) {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        Map<String, Object> 결과 = new HashMap<>();
        String 이슈전환_아이디 = null;

        try {
            Issue 지라이슈 = restClient.getIssueClient().getIssue(이슈_키_또는_아이디).claim();
            이슈전환_아이디 = 이슈전환_아이디_조회하기(서버정보, 이슈_키_또는_아이디, 상태_아이디);

            if (이슈전환_아이디 != null) {
                TransitionInput transitionInput = new TransitionInput(Integer.parseInt(이슈전환_아이디));
                Promise<Void> 변경결과 = restClient.getIssueClient().transition(지라이슈, transitionInput);
                변경결과.claim();

                결과.put("success", true);
                결과.put("message", "이슈 상태 변경 성공");
            }
            else {
                String 에러로그 = "온프레미스 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + 이슈_키_또는_아이디 + ") :: 상태 아이디(" + 상태_아이디 + ") :: 해당 업무 흐름으로 변경이 불가능 합니다.";
                log.error(에러로그);

                결과.put("success", false);
                결과.put("message", "변경할 이슈 상태가 존재하지 않습니다.");
            }
        }
        catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + 이슈_키_또는_아이디 + ") :: 상태 아이디(" + 상태_아이디 + ") :: 전환 아이디(" + 이슈전환_아이디 + ") :: 이슈_상태_변경하기에 실패하였습니다.");

            결과.put("success", false);
            결과.put("message", 에러로그);
        }

        return 결과;
    }

    public String 이슈전환_아이디_조회하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디, String 상태_아이디) {

        try {
            WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

            String endpoint = "/rest/api/2/issue/" + 이슈_키_또는_아이디 + "/transitions";

            지라이슈전환_데이터 이슈전환_데이터 = 지라유틸.get(webClient, endpoint, 지라이슈전환_데이터.class).block();

            return Optional.ofNullable(이슈전환_데이터)
                    .map(지라이슈전환_데이터::getTransitions)
                    .orElse(Collections.emptyList()).stream()
                    .filter(데이터 -> {
                        if (데이터.getTo() != null && !StringUtils.isBlank(데이터.getTo().getId())) {
                            return 상태_아이디.equals(데이터.getTo().getId());
                        }
                        return false;
                    })
                    .findFirst()
                    .map(지라이슈전환_데이터.Transition::getId)
                    .orElse(null);

        } catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + 이슈_키_또는_아이디 + ") :: 상태 아이디(" + 상태_아이디 + ") :: 이슈전환_아이디_조회하기에 실패하였습니다.");
            throw new IllegalArgumentException(ErrorCode.이슈전환_조회_오류.getErrorMsg() + " :: " + 에러로그);
        }
    }

    @Override
    public Map<String, Object> 이슈_삭제하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디) {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        Map<String, Object> 결과 = new HashMap<String, Object>();
        try {
            boolean 하위이슈_삭제유무 = 지라API_정보.getParameter().isDeleteSubtasks();
            Promise<Void> 삭제결과 = restClient.getIssueClient().deleteIssue(이슈_키_또는_아이디, 하위이슈_삭제유무);
            삭제결과.claim();

            결과.put("success", true);
            결과.put("message", "이슈 삭제 성공");
        }
        catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") :: 이슈_삭제하기 중 오류 :: " + 이슈_키_또는_아이디);
            log.error(에러로그);

            결과.put("success", false);
            결과.put("message", 에러로그);
        }

        return 결과;
    }

    public 지라이슈_데이터 지라이슈_데이터로_변환(Issue 지라이슈) {
        return 지라이슈_데이터로_변환(지라이슈, null);
    }

    public 지라이슈_데이터 지라이슈_데이터로_변환(Issue 지라이슈, String 이슈_키_또는_아이디) {
        지라이슈_데이터 지라이슈_데이터 = new 지라이슈_데이터();
        지라이슈필드_데이터 지라이슈필드_데이터 = new 지라이슈필드_데이터();

        지라이슈_데이터.setId(지라이슈.getId().toString());
        지라이슈_데이터.setKey(지라이슈.getKey());
        지라이슈_데이터.setSelf(지라이슈.getSelf().toString());

        try {
            ObjectMapper objectMapper = new ObjectMapper();
            objectMapper.registerModule(new JodaModule());

            String rawData = objectMapper.writeValueAsString(지라이슈);
            지라이슈_데이터.setRawData(rawData);
        }
        catch (JsonProcessingException e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라 :: 지라이슈_데이터로_변환 중 오류 :: " + 이슈_키_또는_아이디);
            log.error(에러로그);
        }

        if (이슈_키_또는_아이디 != null) {
            지라이슈_데이터.setUpperKey(이슈_키_또는_아이디);
        }

        // 필드
        // 초기화
        ProjectDTO 프로젝트 = ProjectDTO.builder().build();
        지라사용자_데이터 보고자 = 지라사용자_데이터.builder().build();
        지라사용자_데이터 담당자 = 지라사용자_데이터.builder().build();

        // 프로젝트
        if (지라이슈.getProject() != null) {

            프로젝트.setSelf(지라이슈.getProject().getSelf().toString());
            프로젝트.setId(지라이슈.getProject().getId().toString());
            프로젝트.setKey(지라이슈.getProject().getKey());
            프로젝트.setName(지라이슈.getProject().getName());

            지라이슈필드_데이터.setProject(프로젝트);
        }

        // 이슈 유형
        if (지라이슈.getIssueType() != null) {

            String 이슈유형_주소 = String.valueOf(지라이슈.getIssueType().getSelf());
            String 이슈유형_아이디 = String.valueOf(지라이슈.getIssueType().getId());
            String 이슈유형_이름 = 지라이슈.getIssueType().getName();
            String 이슈유형_내용 = 지라이슈.getIssueType().getDescription();
            Boolean 이슈유형_서브테스크여부 = 지라이슈.getIssueType().isSubtask();

            IssueTypeDTO 이슈유형 = new IssueTypeDTO();
            이슈유형.setSelf(이슈유형_주소);
            이슈유형.setId(이슈유형_아이디);
            이슈유형.setName(이슈유형_이름);
            이슈유형.setDescription(이슈유형_내용);
            이슈유형.setSubtask(이슈유형_서브테스크여부);

            지라이슈필드_데이터.setIssuetype(이슈유형);
        }

        // 생성자

        // 보고자
        if (지라이슈.getReporter() != null) {

            보고자.setAccountId(지라이슈.getReporter().getName());
            보고자.setEmailAddress(지라이슈.getReporter().getEmailAddress());
            보고자.setDisplayName(지라이슈.getReporter().getDisplayName());

            지라이슈필드_데이터.setReporter(보고자);
        }

        // 담당자
        if (지라이슈.getAssignee() != null) {

            담당자.setAccountId(지라이슈.getAssignee().getName());
            담당자.setEmailAddress(지라이슈.getAssignee().getEmailAddress());
            담당자.setDisplayName(지라이슈.getAssignee().getDisplayName());

            지라이슈필드_데이터.setAssignee(담당자);
        }

        // 라벨
        if (지라이슈.getLabels() != null) {
            Set<String> 라벨_목록 = 지라이슈.getLabels();
            List<String> 이슈라벨 = new ArrayList<>(라벨_목록);
            지라이슈필드_데이터.setLabels(이슈라벨);
        }

        // 우선 순위
        if (지라이슈.getPriority() != null) {

            String 이슈우선순위_주소 = String.valueOf(지라이슈.getPriority().getSelf());
            String 이슈우선순위_아이디 = String.valueOf(지라이슈.getPriority().getId());
            String 이슈우선순위_이름 = 지라이슈.getPriority().getName();

            IssuePriorityDTO 이슈우선순위 = new IssuePriorityDTO();
            이슈우선순위.setSelf(이슈우선순위_주소);
            이슈우선순위.setId(이슈우선순위_아이디);
            이슈우선순위.setName(이슈우선순위_이름);
            // description

            지라이슈필드_데이터.setPriority(이슈우선순위);
        }

        // 상태 값
        if (지라이슈.getStatus() != null) {

            String 이슈상태_주소 = String.valueOf(지라이슈.getStatus().getSelf());
            String 이슈상태_아이디 = String.valueOf(지라이슈.getStatus().getId());
            String 이슈상태_이름 = 지라이슈.getStatus().getName();
            String 이슈상태_설명 =  지라이슈.getStatus().getDescription();

            IssueStatusDTO 이슈상태 = new IssueStatusDTO();
            이슈상태.setSelf(이슈상태_주소);
            이슈상태.setId(이슈상태_아이디);
            이슈상태.setName(이슈상태_이름);
            이슈상태.setDescription(이슈상태_설명);

            지라이슈필드_데이터.setStatus(이슈상태);
        }

        // 해결책
        if (지라이슈.getResolution() != null) {

            String 이슈해결책_주소 = String.valueOf(지라이슈.getResolution().getSelf());
            String 이슈해결책_아이디 = String.valueOf(지라이슈.getResolution().getId());
            String 이슈해결책_이름 = 지라이슈.getResolution().getName();
            String 이슈해결책_설명 = 지라이슈.getResolution().getDescription();

            이슈해결책_데이터 이슈해결책 = new 이슈해결책_데이터();
            이슈해결책.setSelf(이슈해결책_주소);
            이슈해결책.setId(이슈해결책_아이디);
            이슈해결책.setName(이슈해결책_이름);
            이슈해결책.setDescription(이슈해결책_설명);

            지라이슈필드_데이터.setResolution(이슈해결책);
        }

        // resolutiondate
        for (IssueField 필드 : 지라이슈.getFields()) {
            if (필드 != null && !필드.getId().isEmpty() && 필드.getId().equals("resolutiondate")) {
                if (필드.getValue().toString() != null) {
                    지라이슈필드_데이터.setResolutiondate(필드.getValue().toString());
                }
                break;
            }
        }

        // created
        if (지라이슈.getCreationDate() != null) {
            String 이슈생성날짜 = String.valueOf(지라이슈.getCreationDate());
            지라이슈필드_데이터.setCreated(이슈생성날짜);
        }

        if (지라이슈.getUpdateDate() != null) {
            String 이슈수정날짜 = String.valueOf(지라이슈.getUpdateDate());
            지라이슈필드_데이터.setUpdated(이슈수정날짜);
        }

        // worklogs
        // BasicUser 타입에서 이메일 데이터를 받아올 수 없어서 고민 중...
        if (지라이슈.getWorklogs() != null) {

            List<지라이슈워크로그_데이터> 이슈워크로그_목록 = new ArrayList<>();

            Iterable<Worklog> 전체이슈워크로그 = 지라이슈.getWorklogs();

            for (Worklog 워크로그 : 전체이슈워크로그) {
                지라사용자_데이터 작성자 = new 지라사용자_데이터();

                String 이슈워크로그_주소 = 워크로그.getSelf().toString();
                BasicUser 이슈워크로그_작성자 = 워크로그.getAuthor();
                String 이슈워크로그_작성자아이디 = 이슈워크로그_작성자.getName();
                //String 이슈워크로그_작성자이메일 = 이슈워크로그_작성자.getSelf().toString();
                BasicUser 이슈워크로그_수정작성자 = 워크로그.getUpdateAuthor();
                String 이슈워크로그_수정작성자아이디 = 이슈워크로그_수정작성자.getName();
                //String 이슈워크로그_수정작성자이메일 = 이슈워크로그_수정작성자.getSelf().toString();
                String 이슈워크로그_생성날짜 = 워크로그.getCreationDate().toString();
                String 이슈워크로그_수정날짜 = 워크로그.getUpdateDate().toString();
                String 이슈워크로그_시작날짜 = 워크로그.getStartDate().toString();
                String 이슈워크로그_소요시간_포맷 = 시간_포맷(워크로그.getMinutesSpent());
                Integer 이슈워크로그_소요시간 = 워크로그.getMinutesSpent() * 60;
                String[] 이슈워크로그_아이디 = 이슈워크로그_주소.split("/");

                지라이슈워크로그_데이터 이슈워크로그 = new 지라이슈워크로그_데이터();
                이슈워크로그.setSelf(이슈워크로그_주소);

                작성자.setAccountId(이슈워크로그_작성자아이디);
                //작성자.setEmailAddress(이슈워크로그_작성자이메일);
                이슈워크로그.setAuthor(작성자);

                작성자.setAccountId(이슈워크로그_수정작성자아이디);
                //작성자.setEmailAddress(이슈워크로그_수정작성자이메일);
                이슈워크로그.setUpdateAuthor(작성자);

                이슈워크로그.setCreated(이슈워크로그_생성날짜);
                이슈워크로그.setUpdated(이슈워크로그_수정날짜);
                이슈워크로그.setStarted(이슈워크로그_시작날짜);
                이슈워크로그.setTimeSpent(이슈워크로그_소요시간_포맷);
                이슈워크로그.setTimeSpentSeconds(이슈워크로그_소요시간);
                이슈워크로그.setId(이슈워크로그_아이디[이슈워크로그_아이디.length - 1]);
                이슈워크로그.setIssueId(지라이슈_데이터.getId());

                이슈워크로그_목록.add(이슈워크로그);
            }
            지라이슈필드_데이터.setWorklogs(이슈워크로그_목록);
        }

        // timespent
        if (지라이슈.getTimeTracking().getTimeSpentMinutes() != null) {
            Integer 이슈소요시간 = 지라이슈.getTimeTracking().getTimeSpentMinutes() * 60;
            지라이슈필드_데이터.setTimespent(이슈소요시간);
        }

        if (지라이슈.getSummary() != null) {
            지라이슈필드_데이터.setSummary(지라이슈.getSummary());
        }

        지라이슈_데이터.setFields(지라이슈필드_데이터);

        return 지라이슈_데이터;
    }

    public String 시간_포맷(int 분) {

        // 1주 = 5일, 1일 = 8시간, 1시간 = 60분
        int 주 = 분 / (5 * 8 * 60);
        int 일 = (분 % (5 * 8 * 60)) / (8 * 60);
        int 시간 = (분 % (8 * 60)) / 60;
        int 남은시간 = 분 % 60;

        StringBuilder 포맷팅 = new StringBuilder();
        if (주 > 0) {
            포맷팅.append(주).append("w ");
        }
        if (일 > 0) {
            포맷팅.append(일).append("d ");
        }
        if (시간 > 0) {
            포맷팅.append(시간).append("h ");
        }
        if (남은시간 > 0 || (주 == 0 && 일 == 0 && 시간 == 0)) {
            포맷팅.append(남은시간).append("m ");
        }

        return 포맷팅.toString().trim();
    }

    @Override
    public 지라이슈_데이터 이슈_상세정보_가져오기(서버정보_데이터 서버정보, String issueKeyOrId) throws Exception {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        try {
            Issue issue = restClient.getIssueClient().getIssue(issueKeyOrId).claim();
            return 지라이슈_데이터로_변환(issue);
        }
        catch (RestClientException e) {
            ErrorLogUtil.exceptionLogging(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: " + issueKeyOrId + ", 이슈_상세정보_가져오기 중 오류");

            if (e.getStatusCode().isPresent() && e.getStatusCode().or(-1) == HttpStatus.SC_NOT_FOUND) {
                return null;
            }

            throw e;
        }
        catch (Exception e) {
             ErrorLogUtil.exceptionLogging(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: "+ issueKeyOrId + ", 이슈_상세정보_가져오기 중 오류");

            throw e;
        }
    }

    @Override
    public List<지라이슈_데이터> 연결이슈_가져오기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디) {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        Set<String> 이슈아이디_세트 = new HashSet<>();
        이슈아이디_세트.add(이슈_키_또는_아이디);

        List<지라이슈_데이터> 연결이슈_목록 = new ArrayList<>();
        연결이슈_가져오기_재귀(restClient, 서버정보, 이슈_키_또는_아이디, 이슈아이디_세트, 연결이슈_목록);

        return 연결이슈_목록;
    }

    public void 연결이슈_가져오기_재귀(JiraRestClient restClient, 서버정보_데이터 서버정보, String 이슈_키_또는_아이디, Set<String> 이슈아이디_세트, List<지라이슈_데이터> 연결이슈_목록) {

        String jql = "issue in linkedIssues("+이슈_키_또는_아이디+")";

        int startAt = 0;
        int 최대_검색수 = 지라API_정보.getParameter().getMaxResults();
        Set<String> fields = new HashSet<>(Arrays.asList("*all")); // 검색 필드

        // 이슈 건수가 1000이 넘을때 이슈 조회를 위한 처리
        SearchResult searchResult;

        try {
            do {
                searchResult = restClient.getSearchClient()
                        .searchJql(jql, 최대_검색수, startAt, fields)
                        .get();
                for (Issue issue : searchResult.getIssues()) {

                    // 요구사항 이슈인 경우는 미수집
                    지라이슈_데이터 연결이슈_데이터 = 지라이슈_데이터로_변환(issue);
                    연결이슈_데이터.set서버아이디(서버정보.getConnectId());
                    String 조회_아이디 = 연결이슈_데이터.조회조건_아이디();

                    AlmIssueEntity 조회_결과;
                    try {
                        조회_결과 = 지라이슈_저장소.findRecentDocById(조회_아이디);
                    } catch (Exception e) {
                        ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                                "온프레미스 지라(" + 서버정보.getUri() + ") ::  연결이슈_아이디 :: "+ 연결이슈_데이터.getKey() + ", 연결 이슈의 요구사항 이슈 여부 확인 중 오류");
                        continue;
                    }

                    if (조회_결과 != null && 조회_결과.getIsReq() == Boolean.TRUE) {
                        log.info("온프레미스 지라({}) ::  연결이슈_아이디 :: {} :: 요구사항 이슈입니다.", 서버정보.getUri(), 연결이슈_데이터.getKey());
                        continue;
                    }

                    if (!이슈아이디_세트.contains(issue.getKey())) {
                        이슈아이디_세트.add(issue.getKey());
                        연결이슈_목록.add(연결이슈_데이터);
                        연결이슈_가져오기_재귀(restClient, 서버정보, issue.getKey(), 이슈아이디_세트, 연결이슈_목록);
                    }
                }

                startAt += 최대_검색수;

            } while (searchResult.getTotal() > startAt);

        }
        catch (InterruptedException e) {
            Thread.currentThread().interrupt(); // 스레드를 다시 interrupt

            ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: "+ 이슈_키_또는_아이디 + ", 연결이슈_가져오기 중 오류");
        }
        catch (Exception e) {
            ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: "+ 이슈_키_또는_아이디 + ", 연결이슈_가져오기 중 오류");
        }
    }

    @Override
    public List<지라이슈_데이터> 서브테스크_가져오기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디) {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        String jql = "parent="+ 이슈_키_또는_아이디;

        int startAt = 0;
        int 최대_검색수 = 지라API_정보.getParameter().getMaxResults();
        Set<String> fields = new HashSet<>(Arrays.asList("*all")); // 검색 필드

        // 이슈 건수가 1000이 넘을때 이슈 조회를 위한 처리
        List<지라이슈_데이터> 반환할_서브테스크_목록 = new ArrayList<>();
        SearchResult searchResult;
        try {
            do {
                searchResult = restClient.getSearchClient()
                        .searchJql(jql, 최대_검색수, startAt, fields)
                        .get();
                for (Issue issue : searchResult.getIssues()) {
                    지라이슈_데이터 지라이슈_데이터 = 지라이슈_데이터로_변환(issue, 이슈_키_또는_아이디);
                    반환할_서브테스크_목록.add(지라이슈_데이터);
                }

                startAt += 최대_검색수;
            } while (searchResult.getTotal() > startAt);

            return 반환할_서브테스크_목록;
        }
        catch (InterruptedException e) {
            Thread.currentThread().interrupt(); // 스레드를 다시 interrupt

            ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: "+ 이슈_키_또는_아이디 + ", 서브테스크 가져오기 중 오류");
            return null;
        }
        catch (Exception e) {
            ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: "+ 이슈_키_또는_아이디 + ", 서브테스크 가져오기 중 오류");
            return null;
        }
    }


    @Override
    public List<지라이슈_데이터> 증분서브테스크_가져오기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디,String startDate, String endDate) {

        JiraRestClient restClient = 지라유틸.온프레미스_통신기_생성(서버정보.getUri(),
                서버정보.getUserId(),
                서버정보.getPasswordOrToken());

        String subtaskEndpoint = 지라API_정보.getParameter().getJql().getSubtask();
        String subtaskJql = 지라API_정보.이슈키_대체하기(subtaskEndpoint, 이슈_키_또는_아이디);
        String updateJql = 지라API_정보.getParameter().getJql().getUpdated();
        if(startDate!=null&&endDate!=null){
            updateJql = 지라API_정보.getParameter().getJql().getSubtaskManualDate();
            updateJql = updateJql.replace("{수동날짜}","updated > '"+startDate+"' AND updated < '"+getNextDate(endDate)+"'");
        }

        String jql = subtaskJql + " and " + updateJql;

        int startAt = 0;
        int 최대_검색수 = 지라API_정보.getParameter().getMaxResults();

        Set<String> fields = new HashSet<>(Arrays.asList("*all")); // 검색 필드

        // 이슈 건수가 1000이 넘을때 이슈 조회를 위한 처리
        List<지라이슈_데이터> 반환할_서브테스크_목록 = new ArrayList<>();
        SearchResult searchResult;
        try {
            do {
                searchResult = restClient.getSearchClient()
                        .searchJql(jql, 최대_검색수, startAt, fields)
                        .get();
                for (Issue issue : searchResult.getIssues()) {
                    지라이슈_데이터 지라이슈_데이터 = 지라이슈_데이터로_변환(issue, 이슈_키_또는_아이디);
                    반환할_서브테스크_목록.add(지라이슈_데이터);
                }

                startAt += 최대_검색수;
            } while (searchResult.getTotal() > startAt);

            return 반환할_서브테스크_목록;
        }
        catch (InterruptedException e) {
            Thread.currentThread().interrupt(); // 스레드를 다시 interrupt

            ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: "+ 이슈_키_또는_아이디 + ", 증분 서브테스크 가져오기 중 오류");
            return null;
        }
        catch (Exception e) {
            ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "온프레미스 지라(" + 서버정보.getUri() + ") ::  이슈_키_또는_아이디 :: "+ 이슈_키_또는_아이디 + ", 증분 서브테스크 가져오기 중 오류");
            return null;
        }
    }

}
