package com.arms.egovframework.javaservice.esframework.repository;

import com.arms.egovframework.javaservice.esframework.esquery.EsQuery;
import com.arms.egovframework.javaservice.esframework.annotation.Recent;
import com.arms.egovframework.javaservice.esframework.esquery.EsQueryBuilder;
import com.arms.egovframework.javaservice.esframework.esquery.filter.TermsQueryFilter;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.data.annotation.Id;
import org.springframework.data.elasticsearch.NoSuchIndexException;
import org.springframework.data.elasticsearch.core.ElasticsearchOperations;
import org.springframework.data.elasticsearch.core.SearchHit;
import org.springframework.data.elasticsearch.core.SearchHits;
import org.springframework.data.elasticsearch.core.query.NativeSearchQuery;
import org.springframework.data.elasticsearch.core.query.NativeSearchQueryBuilder;
import org.springframework.data.elasticsearch.core.query.Query;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import static com.arms.config.ApplicationContextProvider.getBean;
import static com.arms.egovframework.javaservice.esframework.util.ReflectionUtil.*;
import static java.util.stream.Collectors.*;

@Slf4j
public class RecentFieldConvertor<S,T> {

    private final Map<Object, List<SearchHit<T>>> searchHitsMapList;

    private final ElasticsearchOperations operations;

    private final List<? extends T> entities;

    public RecentFieldConvertor(Iterable<S> entities, Class<T> tClass) throws  IllegalStateException{

        this.entities = StreamSupport.stream(entities.spliterator(), false).map(tClass::cast).collect(toList());

        if(this.entities.isEmpty()){
            throw new IllegalStateException("저장할 데이터가 없습니다.");
        }

        String recentFieldName = fieldInfo(this.entities, Recent.class).getName();

        String idFieldName = fieldInfo(this.entities,Id.class).getName();

        this.operations = getBean(ElasticsearchOperations.class);

        EsQuery esQuery
            = new EsQueryBuilder()
            .bool(
                new TermsQueryFilter(idFieldName, fieldValues(this.entities,Id.class)),
                new TermsQueryFilter(recentFieldName,true)
            );

        BoolQueryBuilder boolQuery = esQuery.getQuery(new ParameterizedTypeReference<>() {
        });

        NativeSearchQuery searchQuery = new NativeSearchQueryBuilder()
            .withQuery(boolQuery)
            .build();

        this.searchHitsMapList = Optional.ofNullable(this.search(searchQuery, tClass))
            .map(schHits ->
                schHits.getSearchHits().stream()
                    .filter(a -> a.getIndex() != null)
                    .sorted(Comparator.comparing(SearchHit::getIndex, Comparator.reverseOrder()))
                    .collect(groupingBy(a -> {
                        try {
                            String fieldName = fieldInfo(a.getContent().getClass(), Id.class).getName();
                            return getValue(a.getContent(),fieldName);
                        } catch (Exception e) {
                            throw new IllegalArgumentException(e);
                        }
                    }))
            ).orElse(null);
    }

    private SearchHits<T> search(Query query,Class<T> clazz) {
        try{
            return operations.search(query, clazz);
        }catch (NoSuchIndexException e){
            String errorMessage = e.getMessage();
            if (errorMessage != null && errorMessage.contains("no such index")) {
                return null;
            }
            throw e;
        }catch (Exception e){
            log.error(e.getMessage());
            throw e;
        }
    }

    public S recentTrue(S newEntity){
        try {
            String name = fieldInfo(newEntity.getClass(), Id.class).getName();

            Object keyObject = getValue(newEntity, name);

            if(searchHitsMapList!=null && searchHitsMapList.containsKey(keyObject)){

                SearchHit<T> searchHit = searchHitsMapList.get(keyObject)
                    .stream()
                    .findFirst()
                    .orElseThrow(() -> new RuntimeException("값이 비워 있습니다."));

                if(newEntity.equals(searchHit.getContent())){
                    return null;
                }else{
                    String fieldName = fieldInfo(newEntity.getClass(), Recent.class).getName();
                    setBoolean(newEntity,fieldName,true);
                    return newEntity;
                }

            }else{
                String fieldName = fieldInfo(newEntity.getClass(), Recent.class).getName();
                setBoolean(newEntity,fieldName,true);
                return newEntity;
            }

        } catch (Exception e) {
            throw new IllegalArgumentException(e);
        }
    }

    private List<SearchHit<T>> recentFalse(Function<Map<Object, List<SearchHit<T>>>,Stream<SearchHit<T>>> function){

        return Optional.ofNullable(searchHitsMapList)
            .map(entitiMap -> function.apply(entitiMap)
                .map(hit -> {
                    try {
                        String fieldName = fieldInfo(hit.getContent().getClass(), Recent.class).getName();
                        setBoolean(hit.getContent(),fieldName,false);
                        return hit;
                    } catch (Exception e) {
                        throw new IllegalArgumentException(e);
                    }
                }).collect(toList())
            )
            .orElse(null);
    }

    public List<SearchHit<T>> setRecentFalseIfNotEqual(T newEntity){

        Function<Map<Object, List<SearchHit<T>>>,Stream<SearchHit<T>>> function = entityMap -> {
            try {
                String fieldName = fieldInfo(newEntity.getClass(), Id.class).getName();
                return entityMap.getOrDefault(getValue(newEntity,fieldName), Collections.emptyList())
                    .stream()
                    .filter(hit -> !newEntity.equals(hit.getContent()));
            } catch (Exception e) {
                throw new IllegalStateException(e);
            }
        };

        return this.recentFalse(function);
    }

    public List<SearchHit<T>> setRecentFalseIfDuplicate(T newEntity){

        Function<Map<Object, List<SearchHit<T>>>,Stream<SearchHit<T>>> function = (entitiMap)-> {
            try {
                String name = fieldInfo(newEntity.getClass(), Id.class).getName();
                Object key = getValue(newEntity,name);
                return entitiMap.getOrDefault(key, Collections.emptyList())
                    .stream()
                    .filter(hit -> newEntity.equals(hit.getContent()))
                    .skip(1);
            } catch (Exception e) {
                throw new IllegalStateException(e);
            }
        };

        return this.recentFalse(function);
    }

    public Map<String, List<T>> listMapGroupByIndex(Function<T, List<SearchHit<T>>> function){
        return entities.stream()
                .filter(Objects::nonNull)
                .map(function)
                .filter(Objects::nonNull)
                .flatMap(searchHits -> searchHits.stream().filter(a->a.getIndex()!=null))
                .collect(groupingBy(SearchHit::getIndex, mapping(SearchHit::getContent, toList())));
    }
}
